from typing import Dict, List, Tuple, Union, Any

import jaydebeapi
import jpype
import json
import logging
import os

from .segestre import Segestre


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class DatabasePlugin(Segestre):
    def __init__(self, *args, **kwargs):

        # Obtain settings from JSON file in the same dir
        path = os.path.abspath(os.path.join(
            os.path.dirname(__file__), "settings.json"))
        with open(path, "r") as f:
            logger.info(f"Reading settings from {path}")
            settings_data = f.read()
            settings = json.loads(settings_data)

        # Initialize segestre and remote plugin base classes
        super().__init__(*args, settings=settings)

        # Boot JVM that will execute queries
        self._start_jvm()

    def build_jdbc_connection_params(self) -> Tuple[str, str, Union[Dict, List]]:
        """Build JDBC connection params for JayDeBeApi.

        This method must be implemented by the inheriting plugin.

        Must return a tuple of three items:
        1. JDBC classname string of the JDBC driver.
        2. Connection URL or DSN.
        3. List of Dict representing connection args.
        """
        raise NotImplementedError()

    def format_query(self, sql: str) -> str:
        """Format the SQL query before executing it.

        This method must be implemented by the inherigin plugin.

        Must return modified or unmodified SQL query that needs to be executed.
        """
        raise NotImplementedError()

    def build_group_and_device_name(self) -> str:
        """Builds unique name to identify custom device"""

        c = self.settings["connection"]

        # Non standard identifier of the custom device name.
        # E.g., database_host instead of database_name.
        device_name = None
        custom_device_identifier_source = c["custom_device_identifier_source"]
        if custom_device_identifier_source not in [
            "Database Name",
            "Host",
            "Host:Port",
            "Custom"
        ]:
            # In case of a wrong identifier, resort to the default
            logger.warning(f"Invalid custom device identifier source '{custom_device_identifier_source}'. Defaulting to 'Database Name'.")
            custom_device_identifier_source = "Database Name"

        if custom_device_identifier_source == "Database Name":
            device_name = c["database_name"]
        elif custom_device_identifier_source == "Host":
            device_name = c["database_host"]
        elif custom_device_identifier_source == "Host:Port":
            device_name = f"{c['database_host']}:{c['database_port']}"
        elif custom_device_identifier_source == "Custom":
            device_name = c["custom_device_identifier"]

        logger.info(f"Custom device identifier selected: '{custom_device_identifier_source}' = '{device_name}'.")

        # Non standard identifier for the custom device group name.
        group_name = None
        custom_device_group_identifier_source = c["custom_device_group_identifier_source"]
        if custom_device_group_identifier_source not in [
            "Database Name",
            "Host",
            "Host:Port",
            "Custom"
        ]:
            logger.info(f"Invalid custom device group identifier source '{custom_device_group_identifier_source}'. Defaulting to 'Host:Port'.")
            custom_device_group_identifier_source = "Host:Port"

        if custom_device_group_identifier_source == "Database Name":
            group_name = c["database_name"]
        elif custom_device_group_identifier_source == "Host":
            group_name = c["database_host"]
        elif custom_device_group_identifier_source == "Host:Port":
            group_name = f"{c['database_host']}:{c['database_port']}"
        elif custom_device_group_identifier_source == "Custom":
            group_name = c["custom_device_group_identifier"]

        logger.info(f"Custom device group identifier selected: '{custom_device_group_identifier_source}' = '{group_name}'.")

        return group_name, device_name

    def _execute_queries(self, queries: List[Tuple[str, Any]]) -> None:
        logger.info("Executing queries")
        connection = cursor = None

        try:
            connection = jaydebeapi.connect(*self.build_jdbc_connection_params())
            logger.info("Connection created")
            cursor = connection.cursor()
            logger.info("Cursor obtained from connection")

            # Substitute which statistics object to collect information from.
            c = self.settings["connection"]
            database_name = c["database_name"]
            object_name = None
            instance_name_sep = "\\"
            # If host contains named instance, then object name equals it
            # e.g. myServerName\myInstanceName
            if instance_name_sep in c["database_host"]:
                object_name = f"MSSQL${c['database_host'].split(instance_name_sep)[-1]}"
            else:
                object_name = "SQLServer"

            for key, q in queries:
                if q["enabled"]:

                    # Substitute instance name filtering if only a single database
                    # is to be monitored.
                    instance_name_condition = (
                        f"instance_name = '{database_name}'"
                        if c["single_database"]
                        else "instance_name != '_Total'"
                    )

                    sql = q["sql"].format(
                        object_name=object_name,
                        instance_name_condition=instance_name_condition
                    )
                    logger.info(f"Executing query '{key}': {sql}")
                    cursor.execute(sql)

                    rs = cursor.fetchall()
                    num_rows_fetched = len(rs) if rs else -1
                    logger.info(f"Rows fetched: {num_rows_fetched}")
                    # https://www.python.org/dev/peps/pep-0249/#description
                    column_names = [col[0] for col in cursor.description]

                    yield key, column_names, rs

                else:
                    logger.info(f"Skipping disabled query: {key}")

        except Exception as e:
            logger.error(e)
            raise e

        finally:
            if cursor:
                cursor.close()
            if connection:
                connection.close()

    def _start_jvm(self) -> None:
        jdbc_jar_name = self.settings["connection"]["jarpath"]
        jdbc_jar_path = os.path.abspath(os.path.join(os.path.dirname(__file__), jdbc_jar_name))
        logger.info(f"Using jdbc jar: {jdbc_jar_path}")

        # Check that JVM exists somewhere. Otherwise we won't be able to
        # run JayDeBeApi
        java_home = os.environ.get("JAVA_HOME", None)
        if not java_home and not self._find_jvm_path():
            raise ConfigException("Unable to find JVM required for JDBC to run")

        # If JAVA_HOME is set, then try to find JVM in standard JAVA_HOME locations.
        # Otherwise search in predefined paths.
        jvm_path = jpype.getDefaultJVMPath() if java_home else self._find_jvm_path()
        jvm_args = f"-Djava.class.path={jdbc_jar_path}"
        logger.info(f"Found JVM at: {jvm_path}")

        try:
            logger.info(f"Attempting to start JVM at: {jvm_path}")
            jpype.startJVM(jvm_path, jvm_args, convertStrings=True)

        except OSError as ose:
            # Do not react to errors of JVM is already started for Jpype
            logger.warning(f"Ignoring OSError JVM already started exception: {ose}")

        except Exception as e:
            logger.error(f"Unknown exception while trying to start JVM: {e}")
            raise ConfigException(f"Could not start JVM: {e}")

    def _find_jvm_path(self):
        paths = [
            "/usr/lib/jvm",
            "../../../gateway/jre/bin/server/jvm.dll",
            "../../../gateway/jre/lib/server/libjvm.so",
            "../../../gateway/jre/lib/amd64/server/libjvm.so",
            "../../../jre/bin/server/jvm.dll",
            "../../../jre/lib/amd64/server/libjvm.so",
            "C:/Program Files/Dynatrace/gateway/jre/bin/server/jvm.dll",
            "/opt/dynatrace/gateway/jre/lib/server/libjvm.so",
            "/opt/dynatrace/gateway/jre/lib/amd64/server/libjvm.so"
        ]

        gateway_home = os.environ.get("GATEWAY_HOME", None)
        if gateway_home:
            paths.append(f"{gateway_home}/jre/bin/server/jvm.dll")
            paths.append(f"{gateway_home}/jre/bin/server/libjvm.so")

        for p in paths:
            if os.path.isfile(p):
                return p

        return None